(** * Miscellaneous code from paper intro *)
Require Import Coq.Bool.Bool.
Require Import Coq.Setoids.Setoid.

(** ** Proof script primer code *)
Goal forall x, x + 0 = 0 + x.
  induction x as [|x IHx].
  all: simpl Nat.add.
  2: rewrite IHx.
  all: reflexivity.
Qed.


Inductive is_even : nat -> Prop :=
| even_O : is_even O
| even_SS : forall x, is_even x -> is_even (S (S x)).

Goal is_even (let x := 100 * 100 * 100 * 100 in
              let y := x * x * x * x in
              y * y * y * y).
Abort.

(** ** Reflective automation primer code *)
Fixpoint check_is_even (n : nat) : bool
  := match n with
     | O => true
     | S O => false
     | S (S n) => check_is_even n
     end.

Theorem soundness (n : nat) : check_is_even n = true -> is_even n.
Proof.
  revert n; fix 1; destruct n as [|[|n]].
  { constructor. }
  { intro H; inversion H. }
  { specialize (soundness n).
    simpl; intro H'; constructor; intuition. }
Qed.

Goal is_even 2000.
  Time repeat (apply even_SS || apply even_O). (* 1.8 s *)
  Undo.
  Time apply soundness; vm_compute; reflexivity. (* 0.004 s *)
Qed.

Goal is_even (10 * 10 * 10 * 10).
  Time apply soundness; vm_compute; reflexivity. (* 0.001 s *)
Qed.

Goal is_even (10 * 10 * 10 * 10 * 10 * 10).
  Time apply soundness; vm_compute; reflexivity. (* 0.044 s *)
Qed.

Goal is_even (10 * 10 * 10 * 10 * 10 * 10 * 10).
  Time apply soundness; vm_compute; reflexivity. (* 0.396 s *)
Qed.

(*
Goal is_even (10 * 10 * 10 * 10 * 10 * 10 * 10 * 10 * 10).
  Time apply soundness; vm_compute; reflexivity. (* 174.322 secs (151.332u,22.468s) *)
Qed.
*)

(** ** Reflective syntax primer code *)
Inductive expr :=
| NatO : expr
| NatS : expr -> expr
| NatMul : expr -> expr -> expr.

Fixpoint denote (t : expr) : nat
  := match t with
     | NatO => O
     | NatS x => S (denote x)
     | NatMul x y => denote x * denote y
     end.

Fixpoint check_is_even_expr (t : expr) : bool
  := match t with
     | NatO => true
     | NatS x => negb (check_is_even_expr x)
     | NatMul x y => orb (check_is_even_expr x) (check_is_even_expr y)
     end.

Notation is_even_or_odd x := ({is_even x} + {~is_even x}).

Lemma is_even_or_odd_x : forall x, (is_even x /\ ~is_even (S x))
                                   \/ (~is_even x /\ is_even (S x)).
Proof.
  induction x as [|x [[IHx0 IHx1]|[IHx0 IHx1]]].
  { left; split; try constructor; intro H; inversion H. }
  { right; split; [ | constructor ]; assumption. }
  { left; split; [ assumption | intro H; inversion_clear H; apply IHx0; assumption ]. }
Qed.

Definition is_even_or_odd_S x (pf : is_even_or_odd x) : is_even_or_odd (S x).
Proof.
  destruct pf as [pf|pf]; [ right | left ];
    abstract (destruct (is_even_or_odd_x x) as [[H0 H1]|[H0 H1]]; tauto).
Defined.

Definition is_even_or_odd_sum
           x y
  : (is_even x /\ is_even y /\ is_even (x + y))
    \/ (~is_even x /\ ~is_even y /\ is_even (x + y))
    \/ (is_even x /\ ~is_even y /\ ~is_even (x + y))
    \/ (~is_even x /\ is_even y /\ ~is_even (x + y)).
Proof.
  revert y; induction x as [|x IHx];
    intro y;
    [ | destruct (is_even_or_odd_x x) as [H|H]; specialize (IHx (S y));
        rewrite <- !plus_n_Sm in IHx ];
    destruct (is_even_or_odd_x y) as [Hy|Hy];
    intuition.
  { left; repeat constructor; assumption. }
  { right; right; left; repeat constructor; assumption. }
Qed.

Definition is_even_or_odd_mul_helper
           x y
  : (is_even x /\ is_even y /\ is_even (x * y))
    \/ (~is_even x /\ is_even y /\ is_even (x * y))
    \/ (is_even x /\ ~is_even y /\ is_even (x * y))
    \/ (~is_even x /\ ~is_even y /\ ~is_even (x * y)).
Proof.
  induction x as [|x IHx]; simpl.
  { destruct (is_even_or_odd_x y); intuition.
    { left; repeat constructor; assumption. }
    { right; right; left; repeat constructor; assumption. } }
  { pose proof (is_even_or_odd_sum y (x * y)).
    pose proof (is_even_or_odd_x x).
    intuition. }
Qed.

Definition is_even_or_odd_mul x y (Hx : is_even_or_odd x) (Hy : is_even_or_odd y)
  : is_even_or_odd (x * y).
Proof.
  destruct Hx, Hy; [ left | left | left | right ];
    abstract (pose proof (is_even_or_odd_mul_helper x y); intuition).
Defined.

Lemma check_is_even_expr_correct (e : expr)
  : check_is_even_expr e = true <-> is_even (denote e).
Proof.
  induction e; simpl in *.
  { repeat constructor. }
  { rewrite negb_true_iff, <- not_true_iff_false, IHe.
    edestruct (is_even_or_odd_x (denote e)); intuition. }
  { rewrite orb_true_iff, IHe1, IHe2.
    edestruct is_even_or_odd_mul_helper;
      intuition solve [ intuition eauto ]. }
Qed.

Theorem check_is_even_expr_sound (e : expr)
  : check_is_even_expr e = true -> is_even (denote e).
Proof. intro; apply check_is_even_expr_correct; assumption. Qed.

Fixpoint nat_to_expr (v : nat) : expr
  := match v with
     | O => NatO
     | S v => NatS (nat_to_expr v)
     end.
Coercion nat_to_expr : nat >-> expr.
Local Infix "*" := NatMul : expr_scope.
Delimit Scope expr_scope with expr.
Open Scope expr_scope.

Set Printing Coercions.
Goal is_even (denote (10 * 10 * 10 * 10 * 10 * 10 * 10 * 10 * 10)%expr).
  Time apply check_is_even_expr_sound; vm_compute; reflexivity. (* 0 s *)
Qed.

Goal is_even (denote (let x0 := 100 * 100 * 100 * 100 in
                      let x1 := x0 * x0 * x0 * x0 in
                      let x2 := x1 * x1 * x1 * x1 in
                      x2)).
  Time apply check_is_even_expr_sound; vm_compute; reflexivity. (* 0.001 s *)
Qed.

Goal is_even (denote (let x0 := 10 * 10 in
                      let x1 := x0 * x0 in
                      let x2 := x1 * x1 in
                      let x3 := x2 * x2 in
                      let x4 := x3 * x3 in
                      let x5 := x4 * x4 in
                      let x6 := x5 * x5 in
                      let x7 := x6 * x6 in
                      let x8 := x7 * x7 in
                      let x9 := x8 * x8 in
                      let x10 := x9 * x9 in
                      let x11 := x10 * x10 in
                      let x12 := x11 * x11 in
                      let x13 := x12 * x12 in
                      let x14 := x13 * x13 in
                      let x15 := x14 * x14 in
                      let x16 := x15 * x15 in
                      let x17 := x16 * x16 in
                      let x18 := x17 * x17 in
                      let x19 := x18 * x18 in
                      let x20 := x19 * x19 in
                      let x21 := x20 * x20 in
                      let x22 := x21 * x21 in
                      let x23 := x22 * x22 in
                      let x24 := x23 * x23 in
                      x24)).
  Time apply check_is_even_expr_sound; vm_compute; reflexivity. (* 19.13 s *)
Qed.

(** The expr type with let ... in ... is in PHOAS.v *)
