(** * Reification by Mtac2 *)
Require Import Coq.Strings.String Coq.ZArith.ZArith.
Require Import Reify.ReifyCommon.
Require Import Mtac2.Mtac2.
Import M.notations.

(** Points of note:

    - We use %\verb|=n>|% to avoid unnecessary normalization /
      reduction

    - We handle fresh binder names manually rather than invoking
      [M.fresh_binder_name], which produces a string with length
      linear in the number of times it has been called so far, rather
      than logarithmic. *)

Module var_context.
  Inductive var_context {var : Type} :=
  | nil
  | cons (n : nat) (v : var) (xs : var_context).
End var_context.

Fixpoint string_of_pos_bin' (p : positive) (rest : string) : string
  := match p with
     | xI x => string_of_pos_bin' x (String "1" rest)
     | xO x => string_of_pos_bin' x (String "0" rest)
     | xH => String "0" (String "b" (String "1" rest))
     end.
Definition string_of_pos_bin (p : positive) : string
  := string_of_pos_bin' p EmptyString.

(** We'd like to just use [M.fresh_binder_name f], but this incurs
    significant overhead (about 2X slower at 27 invocations) because
    it has strings linear in the number of repeated invocations,
    rather than logarithmic, so instead we handle binder names
    manually. *)

(** [norm_string] is useful for printing / debugging, but incurs a bit
    of overhead. *)

Definition norm_string (v : string) : M string :=
  (mfix1 norm_string (s : string) : M string :=
     (mmatch s with
     | EmptyString => M.ret EmptyString
     | [? a b] (String a b) => b' <- norm_string b; M.ret (String a b')
      end)) v.
Definition name_binder {A B} (f : A -> B) (var_idx : positive) : M string
  := M.ret (String "x" (string_of_pos_bin var_idx)).
Definition find_in_ctx {var : Type} (term : nat)
           (ctx : @var_context.var_context var)
  : M (option var)
  := (mfix1 find_in_ctx (ctx : @var_context.var_context var) : M (option var) :=
        (mmatch ctx with
        | [? v xs] (var_context.cons term v xs)
          =n> M.ret (Some v)
        | [? x v xs] (var_context.cons x v xs)
          =n> find_in_ctx xs
        | _ => M.ret None
         end)) ctx.
Definition reify_helper {var : Type} (term : nat)
           (ctx : @var_context.var_context var)
  : M (@expr var)
  := ((mfix3 reify_helper (term : nat) (var_idx : positive)
             (ctx : @var_context.var_context var)
       : M (@expr var) :=
         lvar <- find_in_ctx term ctx;
           match lvar with
           | Some v => M.ret (@Var var v)
           | None
             =>
             (mmatch term with
             | O
               =n> M.ret (@NatO var)
             | [? x] (S x)
               =n> (rx <- reify_helper x var_idx ctx;
                      M.ret (@NatS var rx))
             | [? x y] (x * y)
               =n> (rx <- reify_helper x var_idx ctx;
                      ry <- reify_helper y var_idx ctx;
                      M.ret (@NatMul var rx ry))
             | [? v f] (@Let_In nat nat v f)
               =n> (rv <- reify_helper v var_idx ctx;
                    x <- name_binder f var_idx;
                    let vx := String.append "var_" x in
                    rf <- (M.nu
                             x mNone
                             (fun x : nat
                              => M.nu
                                   vx mNone
                                   (fun vx : var
                                    => let fx := reduce (RedWhd [rl:RedBeta]) (f x) in
                                       rf <- (reify_helper
                                                fx
                                                (Pos.succ var_idx)
                                                (var_context.cons x vx ctx));
                                       M.abs_fun vx rf)));
                    M.ret (@LetIn var rv rf))
              end)
           end) term 1%positive ctx).
Definition reify (var : Type) (term : nat) : M (@expr var)
  := reify_helper term var_context.nil.
Definition Reify (term : nat) : M Expr
  := \nu var:Type, r <- reify var term; M.abs_fun var r.

Ltac Reify' x := constr:(ltac:(mrun (@Reify x))).
Ltac Reify x := Reify' x.
Ltac do_Reify_rhs _ := do_Reify_rhs_of ltac:(Reify) ().
Ltac post_Reify_rhs _ := ReifyCommon.post_Reify_rhs ().
Ltac Reify_rhs _ := Reify_rhs_of ltac:(Reify) ().
